from perceiver.data.text import (
    BookCorpusDataModule,
    BookCorpusOpenDataModule,
    C4DataModule,
    Enwik8DataModule,
    ImdbDataModule,
    WikipediaDataModule,
    WikiTextDataModule,
)
